@@ -1,18 +1,22 @@
+/* -*- mode: C; mode: fold; -*- */
 /*
  *  Copyright (c) 1992, 1995 John E. Davis  (davis@space.mit.edu)
  *  All Rights Reserved.
  */
 
+#include "config.h"
+#include "jed-feat.h"
+/*{{{ Include Files */
+
 #include <stdio.h>
 #include <setjmp.h>
 #if defined (msdos) || defined (__os2_16__)
-#include <process.h>
-#include <dos.h>
+# include <process.h>
+# include <dos.h>
 #endif
 #include <limits.h>
 #include <string.h>
 
-#include "config.h"
 #include "buffer.h"
 #include "window.h"
 #include "file.h"
@@ -20,24 +24,33 @@
 #include "misc.h"
 #include "paste.h"
 #include "sysdep.h"
+#include "cmds.h"
 
-extern int Goal_Column;
+#if JED_HAS_SUBPROCESSES
+#include "jprocess.h"
+#endif
 
-typedef struct
+
+/*}}}*/
+
+typedef struct /*{{{*/
 {
    jmp_buf b;
-} jmp_buf_struct;
+}
+
+/*}}}*/
+jmp_buf_struct;
 
 extern jmp_buf_struct Jump_Buffer, *Jump_Buffer_Ptr;
 
+typedef struct Buffer_List_Type /*{{{*/
+{
+   Buffer *buf;
+   struct Buffer_List_Type *next;
+}
 
-typedef struct Buffer_List_Type
-  {
-      Buffer *buf;
-      struct Buffer_List_Type *next;
-  }
+/*}}}*/
 Buffer_List_Type;
-
 Buffer_List_Type *Buffer_Stack;   /* a filo */
 
 #ifndef ULONG_MAX
@@ -45,11 +58,15 @@ Buffer_List_Type *Buffer_Stack;   /* a filo */
 #else
 #define MAX_LONG ULONG_MAX
 #endif
- 
 
 Buffer *CBuf;
 Line *CLine;
-Buffer_Local_Type Buffer_Local = {8, 0};
+Buffer_Local_Type Buffer_Local = /*{{{*/
+{
+   8, 0
+};
+
+/*}}}*/
 
 int Number_One = 1;		       /* these should be const but  */
 int Number_Zero = 0;		       /* some compilers complain */
@@ -62,7 +79,7 @@ unsigned int Max_LineNum;
 int Point = 0;
 
 /* move point to top of buffer */
-int bob()
+int bob() /*{{{*/
 {
    CLine = CBuf->beg;
    Point = 0;
@@ -70,7 +87,9 @@ int bob()
    return(1);
 }
 
-int eob()                      /* point to end of buffer */
+/*}}}*/
+
+int eob()                      /* point to end of buffer */ /*{{{*/
 {
    CLine = CBuf->end;
    LineNum = Max_LineNum;
@@ -79,54 +98,62 @@ int eob()                      /* point to end of buffer */
 	Point = 0;
 	return(0);
      }
-
-    Point = CLine->len - 1;
-    if (Point < 0)
-      {
-          Point = 0;
-          return(1);
-      }
-
-    if ((CLine->data[Point] != '\n') || (CBuf == MiniBuffer)) Point++;
-    return(1);
+   
+   Point = CLine->len - 1;
+   if (Point < 0)
+     {
+	Point = 0;
+	return(1);
+     }
+   
+   if ((CLine->data[Point] != '\n') || (CBuf == MiniBuffer)) Point++;
+   return(1);
 }
 
-int bol()
+/*}}}*/
+
+int bol() /*{{{*/
 {
-    Point = 0;
-    Goal_Column = 1;
-    return(1);
+   Point = 0;
+   Goal_Column = 1;
+   return(1);
 }
 
-int eol()
+/*}}}*/
+
+int eol (void) /*{{{*/
 {
-    int newline = 1;
-    if (CLine != NULL)
-      {
-          Point = CLine->len - 1;
-          if (Point < 0)
-            {
-                Point = 0;
-                return(0);
-            }
-
-          if (((CLine->data)[Point] != '\n') || (CBuf == MiniBuffer)) newline = 0;
-          if (!newline) Point++;
-      }
-
-    else Point = 0;
-    return(1);
+   int nl = 1;
+   if (CLine != NULL)
+     {
+	Point = CLine->len - 1;
+	if (Point < 0)
+	  {
+	     Point = 0;
+	     return 0;
+	  }
+	
+	if (((CLine->data)[Point] != '\n') || (CBuf == MiniBuffer)) nl = 0;
+	if (!nl) Point++;
+     }
+   
+   else Point = 0;
+   return 1;
 }
 
-int bobp()                /* returns 1 if top line of buffer, 0 otherwise */
+/*}}}*/
+
+int bobp()                /* returns 1 if top line of buffer, 0 otherwise */ /*{{{*/
 {
-    if ((CBuf->beg == CLine) && (Point == 0)) return(1); else return(0);
+   if ((CBuf->beg == CLine) && (Point == 0)) return(1); else return(0);
 }
 
-int eolp()
+/*}}}*/
+
+int eolp() /*{{{*/
 {
    int len;
-
+   
    if (CLine == NULL) return(1);
    len = CLine->len;
    if (len == 0) return(1);
@@ -137,74 +164,84 @@ int eolp()
 	if (Point < len) return 0; else return (1);
      }
    
-
-    /* Point is either len or len - 1 */
-
+   
+   /* Point is either len or len - 1 */
+   
    if ((CLine->data)[len-1] == '\n') return (1);
-
+   
    if (Point == len) return(1);
    Point = len - 1;
    return(0);
 }
 
-int eobp()
+/*}}}*/
+
+int eobp() /*{{{*/
 {
-    if (CLine != CBuf->end) return(0);
-    return(eolp());
+   if (CLine != CBuf->end) return(0);
+   return(eolp());
 }
 
-int bolp()
+/*}}}*/
+
+int bolp() /*{{{*/
 {
-    if (Point) return(0); else return(1);
+   if (Point) return(0); else return(1);
 }
 
+/*}}}*/
+
 /*  Attempt to goback n lines, return actual number. */
-int prevline(int *n)
+int prevline(int *n) /*{{{*/
 {
-    int i = 0;
-    Line *prev;
-
-    Point = 0;     /* return to beginning of this line */
-    while (i < *n)
-      {
-          prev = CLine->prev;
-          while((prev != NULL) && (prev->len == 0)) prev = prev->prev;
-          if (prev == NULL) break;
-          i++;
-          CLine = prev;
-      }
-
+   int i = 0;
+   Line *prev;
+   
+   Point = 0;     /* return to beginning of this line */
+   while (i < *n)
+     {
+	prev = CLine->prev;
+	while((prev != NULL) && (prev->len == 0)) prev = prev->prev;
+	if (prev == NULL) break;
+	i++;
+	CLine = prev;
+     }
+   
    if (i) eol();
    LineNum -= (unsigned int) i;
    return(i);
 }
 
-int nextline(int *np)
+/*}}}*/
+
+int nextline(int *np) /*{{{*/
 {
    register Line *next;
    register int i = 0, n = *np;
-
-    while(i < n)
-      {
-	 next = CLine->next;
-	 if (next == NULL) break;
-	 CLine = next;
-	 i++;
-      }
-
+   
+   while(i < n)
+     {
+	next = CLine->next;
+	if (next == NULL) break;
+	CLine = next;
+	i++;
+     }
+   
    if (i) Point = 0;
    LineNum += (unsigned int) i;
    return(i);
 }
 
+/*}}}*/
+
 /* The algorithm:  go to Point = len - 1.  Then forward a line counted as 1 */
-int forwchars(int *np)
+int forwchars(int *np) /*{{{*/
 {
    int len, total = 0;
    unsigned char *p;
    Line *next;
    int n = *np;
-
+   
    if (n < 0) return 0;
    if (CBuf == MiniBuffer)
      {
@@ -213,39 +250,41 @@ int forwchars(int *np)
 	Point += n;
 	return n;
      }
+   
+   while(1)
+     {
+	len = CLine->len;
+	/* if (Point == len) return(total); */
+	p = CLine->data + Point;
+	while(n && (Point < len) && (*p != '\n'))
+	  {
+	     Point++;
+	     n--;
+	     p++;
+	     total++;
+	  }
+	if (!n) return(total);
+	if ((Point < len) && (*p == '\n'))
+	  {
+	     total++;
+	     n--;
+	  }
 	
-    while(1)
-      {
-          len = CLine->len;
-          /* if (Point == len) return(total); */
-          p = CLine->data + Point;
-          while(n && (Point < len) && (*p != '\n'))
-            {
-                Point++;
-                n--;
-                p++;
-                total++;
-            }
-	 if (!n) return(total);
-	 if ((Point < len) && (*p == '\n'))
-            {
-                total++;
-                n--;
-            }
-
-          if (NULL == (next = CLine->next))
-            {
-	       if ((Point < len) && (*p == '\n')) total--;
-	       return(total);
-            }
-
-          Point = 0;
-          CLine = next;
-	 LineNum++;
-      }
+	if (NULL == (next = CLine->next))
+	  {
+	     if ((Point < len) && (*p == '\n')) total--;
+	     return(total);
+	  }
+	
+	Point = 0;
+	CLine = next;
+	LineNum++;
+     }
 }
 
-int backwchars(int *np)
+/*}}}*/
+
+int backwchars(int *np) /*{{{*/
 {
    int total = 0;
    int n = *np;
@@ -258,50 +297,62 @@ int backwchars(int *np)
 	return (n);
      }
    
-
-    do
-      {
-          if (n <= Point)
-            {
-                total += n;
-                Point = Point - n;
-                return(total);
-            }
-          n = n - Point - 1;   /* Point + newline */
-          total += Point + 1;   /* ditto */
-      }
-    while(prevline(&Number_One));
-
-    return(total - 1);
+   
+   do
+     {
+	if (n <= Point)
+	  {
+	     total += n;
+	     Point = Point - n;
+	     return(total);
+	  }
+	n = n - Point - 1;   /* Point + newline */
+	total += Point + 1;   /* ditto */
+     }
+   while(prevline(&Number_One));
+   
+   return(total - 1);
 }
+
+/*}}}*/
+
 /* assuming 8 bit bytes */
 #define BUNCH_SIZE 8 * sizeof(long)
-static unsigned char NewLine_Buffer[1] = {'\n'};
-typedef struct Bunch_Lines_Type
+static unsigned char NewLine_Buffer[1] = /*{{{*/
+{
+   '\n'
+};
+
+/*}}}*/
+
+typedef struct Bunch_Lines_Type /*{{{*/
 {
    struct Bunch_Lines_Type *next;
    unsigned long flags;			       /* describes which are free */
    Line lines[BUNCH_SIZE];
-} Bunch_Lines_Type;
+}
 
-static int Next_Free_Offset = BUNCH_SIZE;
-static int Last_Free_Offset;
+/*}}}*/
+Bunch_Lines_Type;
+
+static unsigned int Next_Free_Offset = BUNCH_SIZE;
+static unsigned int Last_Free_Offset;
 static Bunch_Lines_Type *Last_Free_Group;
 static Bunch_Lines_Type *Bunch_Lines;
 static unsigned int Number_Freed;      /* amount of Lines available */
 
-static Line *create_line_from_bunch(void)
+static Line *create_line_from_bunch(void) /*{{{*/
 {
    register Bunch_Lines_Type *b, *bsave;
    Line *l;
    unsigned long flags;
    int count;
    
-   if (Last_Free_Group != NULL) 
+   if (Last_Free_Group != NULL)
      {
 	l = &Last_Free_Group->lines[Last_Free_Offset];
 	flags = ((unsigned long) 1L << Last_Free_Offset);
-	if ((Last_Free_Group->flags & flags) == 0) 
+	if ((Last_Free_Group->flags & flags) == 0)
 	  {
 	     exit_error("create group: internal error 1", 1);
 	  }
@@ -315,7 +366,7 @@ static Line *create_line_from_bunch(void)
    if (Next_Free_Offset < BUNCH_SIZE)
      {
 	flags = ((unsigned long) 1L << Next_Free_Offset);
-	if ((Bunch_Lines->flags & flags) == 0) 
+	if ((Bunch_Lines->flags & flags) == 0)
 	  {
 	     exit_error("free group: internal error 2", 1);
 	  }
@@ -342,17 +393,17 @@ static Line *create_line_from_bunch(void)
 	       {
 		  flags = b->flags;
 		  count = 0;
-		  while ((flags & 1) == 0) 
+		  while ((flags & 1) == 0)
 		    {
 		       flags = flags >> 1;
 		       count++;
 		    }
 		  l = &b->lines[count];
 		  flags = (unsigned long) 1 << count;
-		  if ((b->flags & flags) == 0) 
+		  if ((b->flags & flags) == 0)
 		    {
-		            exit_error("free group: internal error 2", 1);
-		    } 
+		       exit_error("free group: internal error 2", 1);
+		    }
 		  b->flags &= ~flags;
 		  Number_Freed--;
 		  return (l);
@@ -364,31 +415,32 @@ static Line *create_line_from_bunch(void)
    
    /* failed so now malloc new bunch */
    
-   if (NULL != (b = (Bunch_Lines_Type *) SLMALLOC(sizeof(Bunch_Lines_Type))))
-     {
-	if (Bunch_Lines == NULL)
-	  {
-	     Bunch_Lines = b;
-	     b->next = b;
-	  }
-	else
-	  {
-	     b->next = Bunch_Lines->next;
-	     Bunch_Lines->next = b;
-	     Bunch_Lines = b;
-	  }
-
-	b->flags = MAX_LONG;
-	Next_Free_Offset = 1;
-	b->flags &= ~(unsigned long) 1;
-	Number_Freed += BUNCH_SIZE - 1;
-	return(&b->lines[0]);
-     }
-   return(NULL);
-}
+   if (NULL == (b = (Bunch_Lines_Type *) jed_malloc (sizeof(Bunch_Lines_Type))))
+     return NULL;
    
+   if (Bunch_Lines == NULL)
+     {
+	Bunch_Lines = b;
+	b->next = b;
+     }
+   else
+     {
+	b->next = Bunch_Lines->next;
+	Bunch_Lines->next = b;
+	Bunch_Lines = b;
+     }
+   
+   b->flags = MAX_LONG;
+   Next_Free_Offset = 1;
+   b->flags &= ~(unsigned long) 1;
+   Number_Freed += BUNCH_SIZE - 1;
+   return(&b->lines[0]);
+}
 
-static void destroy_bunch_line(Line *l)
+/*}}}*/
+
+
+static void destroy_bunch_line(Line *l) /*{{{*/
 {
 #ifdef _MSC_VER
    Line *ll;
@@ -398,7 +450,7 @@ static void destroy_bunch_line(Line *l)
    
    register Bunch_Lines_Type *b, *bsave, *last, *next;
    static Bunch_Lines_Type *slast;
-
+   
    if (slast != NULL) last = slast;
    else last = Bunch_Lines;
    
@@ -408,16 +460,16 @@ static void destroy_bunch_line(Line *l)
      {
 	ll = b->lines;
 	if (
-#if (defined (msdos) && !defined(__WATCOMC__)) || defined (__os2_16__)
-	     /* stupid DOS and its memory segmentation forces me to consider
-	        segment then offset */
+#if (defined (msdos) && !defined(__WATCOMC__) && !defined(__WIN32__)) || defined (__os2_16__)
+	    /* stupid DOS and its memory segmentation forces me to consider
+	     segment then offset */
 	    (FP_SEG(ll) == FP_SEG(l)) &&
 #endif
 	    ((ll <= l) && (l < ll + BUNCH_SIZE)))
 	  {
-	     while (ll != l) ll++;
-	     Last_Free_Offset = (int) (l - b->lines);
-	     if (b->flags & (1L << Last_Free_Offset)) 
+	     Last_Free_Offset = (unsigned int) (l - ll);
+
+	     if (b->flags & (1L << Last_Free_Offset))
 	       {
 		  exit_error("free group: internal error 2", 1);
 	       }
@@ -433,7 +485,7 @@ static void destroy_bunch_line(Line *l)
 		    }
 		  
 		  last->next = b->next;
-
+		  
 		  if (b == Bunch_Lines)
 		    {
 		       if (last == b)
@@ -443,7 +495,7 @@ static void destroy_bunch_line(Line *l)
 			     last = last->next;
 			     if (last == b) last = NULL; */
 			 }
-
+		       
 		       Bunch_Lines = last;
 		       Next_Free_Offset = BUNCH_SIZE;
 		    }
@@ -455,7 +507,7 @@ static void destroy_bunch_line(Line *l)
 		  Number_Freed -= BUNCH_SIZE - 1;
 		  slast = last;
 	       }
-	     else 
+	     else
 	       {
 		  Number_Freed++;
 	       }
@@ -472,35 +524,35 @@ static void destroy_bunch_line(Line *l)
    exit_error("destroy_bunch_line: internal error 1", 1);
 }
 
-Line *make_line1(int size)
+/*}}}*/
+
+Line *make_line1(int size) /*{{{*/
 {
    Line *new_line;
    unsigned char *data = NULL;
-   char buff[80];
    unsigned int chunk;
-
+   
    /* 4 byte chunks */
-#ifdef msdos
-   chunk = (unsigned) (size + 3) & 0xFFFC;
+#if defined(msdos) && !defined(__WIN32__)
+   chunk = (unsigned) (size + 3) & 0xFFFCU;
 #else
-   chunk = ((unsigned) (size + 3)) & 0xFFFFFFFC; 
+   chunk = ((unsigned) (size + 3)) & 0xFFFFFFFCU;
 #endif
    new_line = (Line *) create_line_from_bunch();
-   if (new_line != NULL) 
+   if (new_line != NULL)
      {
-	if (size == 1) 
+	if (size == 1)
 	  {
 	     data = NewLine_Buffer;
 	     chunk = 1;
 	  }
-	else data = (unsigned char *) SLMALLOC(chunk);   /* was chunk + 1 */
+	else data = (unsigned char *) jed_malloc (chunk);   /* was chunk + 1 */
      }
-
+   
    if ((new_line == NULL) || (data == NULL))
      {
 	*Error_Buffer = 0;	       /* this is critical */
-	sprintf(buff, "Malloc Error in make_line: requested size: %d.", size);
-	msg_error(buff);
+	jed_verror ("Malloc Error in make_line: requested size: %d.", size);
 	longjmp(Jump_Buffer_Ptr->b, 1);
 	/* exit_error(buff); */
      }
@@ -508,39 +560,44 @@ Line *make_line1(int size)
    new_line->len = 0;
 #ifdef KEEP_SPACE_INFO
    new_line->space = chunk;
+#endif
+#if JED_HAS_LINE_ATTRIBUTES
+   new_line->flags = 0;
 #endif
    return(new_line);
- }
+}
+
+/*}}}*/
 
 /* adds a new link to list of lines at current point */
-unsigned char *make_line(int size)
+unsigned char *make_line(int size) /*{{{*/
 {
-    Line *new_line;
-
-    new_line = make_line1(size);
-    /* if CLine is Null, then we are at the top of a NEW buffer.  Make this
-       explicit. */
-    if (CLine == NULL)
-      {
-          new_line -> prev = NULL;
-          new_line -> next = NULL;
-          CBuf -> beg = CBuf ->end = new_line;
-      }
-    else if (CLine == CBuf->end) /* at end of buffer */
-      {
-          CBuf->end  = new_line;
-          new_line->next = NULL;
-          new_line->prev = CLine;
-          CLine->next = new_line;
-      }
-    else
-      {
-          new_line -> next = CLine -> next;
-          if (CLine->next != NULL) CLine->next->prev = new_line;
-          CLine->next = new_line;
-          new_line->prev = CLine;
-      }
-
+   Line *new_line;
+   
+   new_line = make_line1(size);
+   /* if CLine is Null, then we are at the top of a NEW buffer.  Make this
+    explicit. */
+   if (CLine == NULL)
+     {
+	new_line -> prev = NULL;
+	new_line -> next = NULL;
+	CBuf -> beg = CBuf ->end = new_line;
+     }
+   else if (CLine == CBuf->end) /* at end of buffer */
+     {
+	CBuf->end  = new_line;
+	new_line->next = NULL;
+	new_line->prev = CLine;
+	CLine->next = new_line;
+     }
+   else
+     {
+	new_line -> next = CLine -> next;
+	if (CLine->next != NULL) CLine->next->prev = new_line;
+	CLine->next = new_line;
+	new_line->prev = CLine;
+     }
+   
    if (CLine == NULL)
      {
 	Max_LineNum = LineNum = 1;
@@ -555,7 +612,9 @@ unsigned char *make_line(int size)
    return(CLine->data);
 }
 
-void free_line(Line *line)
+/*}}}*/
+
+void free_line(Line *line) /*{{{*/
 {
    register unsigned char *dat = line->data;
    
@@ -563,44 +622,47 @@ void free_line(Line *line)
    destroy_bunch_line(line);
 }
 
+/*}}}*/
+
 /* deletes the line we are on and returns the prev one.  It does not
  * delete the top line of the buffer.   Furthermore, it does not
  *  update any marks.  */
 
-int delete_line()
+int delete_line() /*{{{*/
 {
-    Line *n, *p, *tthis;
-
-    p = CLine -> prev;
-    if (p == NULL) return(1);
-
-    n = CLine -> next;
-    tthis = CLine;
-    if (n == NULL)
-      {
-          CBuf->end = p;
-          p->next = NULL;
-      }
-    else
-      {
-          p->next = n;
-          n->prev = p;
-      }
-
-    free_line(tthis);
-    CLine = p;
+   Line *n, *p, *tthis;
+   
+   p = CLine -> prev;
+   if (p == NULL) return(1);
+   
+   n = CLine -> next;
+   tthis = CLine;
+   if (n == NULL)
+     {
+	CBuf->end = p;
+	p->next = NULL;
+     }
+   else
+     {
+	p->next = n;
+	n->prev = p;
+     }
+   
+   free_line(tthis);
+   CLine = p;
    LineNum--;
    Max_LineNum--;
    
    return(0);
 }
 
-unsigned char *remake_line(int size)
+/*}}}*/
+
+unsigned char *remake_line(int size) /*{{{*/
 {
-   char buf[80];
    unsigned char *d = CLine->data;
    unsigned int mask;
-#ifdef msdos
+#if defined(msdos) && !defined(__WIN32__)
    mask = 0xFFFCu;
 #else
    mask = 0xFFFFFFFCu;
@@ -609,27 +671,26 @@ unsigned char *remake_line(int size)
    
    if (d == NewLine_Buffer)
      {
-	if (NULL != (d = (unsigned char *) SLMALLOC(size))) *d = '\n';
+	if (NULL != (d = (unsigned char *) jed_malloc (size))) *d = '\n';
      }
-   else 
+   else
      {
 #if 0
 #ifndef KEEP_SPACE_INFO
 	if (((CLine->len + 3) & mask) == size) return d;
 #endif
 #endif
-	d = (unsigned char *) SLREALLOC(d, size);
+	d = (unsigned char *) jed_realloc ((char *) d, size);
      }
-	     
+   
    if (d == NULL)
      {
 	*Error_Buffer = 0;	       /* critical error */
-	sprintf(buf, "remake_line: realloc error!, size = %d", size);
-	msg_error(buf);
+	jed_verror ("remake_line: realloc error!, size = %d", size);
 	longjmp(Jump_Buffer_Ptr->b, 1);
 	/* exit_error(buf); */
      }
-
+   
 #ifdef KEEP_SPACE_INFO
    CLine->space = size;
 #endif
@@ -637,52 +698,55 @@ unsigned char *remake_line(int size)
    return(d);
 }
 
-void uniquely_name_buffer(char *trry)
-{
-    Buffer *b;
-    int exists, version = 0, n;
-    char neew[48];
+/*}}}*/
 
-    *CBuf->name = 0;
-    strncpy(neew, trry, 45);  neew[45] = 0;
+void uniquely_name_buffer(char *trry) /*{{{*/
+{
+   Buffer *b;
+   int exists, version = 0, n;
+   char neew[48];
+   
+   *CBuf->name = 0;
+   strncpy(neew, trry, 45);  neew[45] = 0;
    n = strlen(neew);
-    while (1)
-      {
-	  exists = 0;
-	  b = CBuf->next;
-	  while(b != CBuf)
-	    {
-		if (!strcmp(neew, b->name)) exists = 1;
-		b = b->next;
-	    }
-	  if (!exists)
-	    {
-		strcpy(CBuf->name,neew);
-		return;
-	    }
-	  version++;
-	  sprintf(neew + n, "<%d>", version);
-      }
+   while (1)
+     {
+	exists = 0;
+	b = CBuf->next;
+	while(b != CBuf)
+	  {
+	     if (!strcmp(neew, b->name)) exists = 1;
+	     b = b->next;
+	  }
+	if (!exists)
+	  {
+	     strcpy(CBuf->name,neew);
+	     return;
+	  }
+	version++;
+	sprintf(neew + n, "<%d>", version);
+     }
 }
 
+/*}}}*/
+
 /* make a buffer and insert it in the list */
-Buffer *make_buffer()
+Buffer *make_buffer (void) /*{{{*/
 {
-    Buffer *newB;
-
-   /* SLCALLOC used since it zeros region too */
-    newB = (Buffer *) SLCALLOC(sizeof(Buffer), 1);
-    if (newB == (Buffer *) NULL)
-      {
-          exit_error("make_buffer: malloc error", 0);
-      }
-
-   MEMSET ((char *) newB, sizeof(Buffer), 0);
+   Buffer *newB;
+   
+   newB = (Buffer *) jed_malloc (sizeof(Buffer));
+   if (newB == NULL)
+     {
+	exit_error("make_buffer: malloc error", 0);
+     }
+   
+   SLMEMSET ((char *) newB, 0, sizeof(Buffer));
    
    newB->keymap = Global_Map;
    newB->c_time = sys_time();
-   newB->tab = User_Vars.tab;
-#ifdef HAS_ABBREVS
+   newB->local_vars.tab = Jed_Tab_Default;
+#if JED_HAS_ABBREVS
    newB->abbrev_table_handle = -1;
 #endif
    
@@ -698,7 +762,7 @@ Buffer *make_buffer()
 	CBuf->prev->next = newB;
 	CBuf->prev = newB;
      }
-
+   
 #ifdef pc_system
    newB->flags |= ADD_CR_ON_WRITE_FLAG;
 #endif
@@ -706,59 +770,70 @@ Buffer *make_buffer()
    return(newB);
 }
 
+/*}}}*/
+
 /* if there is a window attached to this buffer, then there are problems
  *  if we get to update() without attaching another one to it.  So
  *  beware! Always make sure CBuf is set too!   kill_buffer command
  *  takes care of this */
-void delete_buffer(Buffer *buf)
+void delete_buffer(Buffer *buf) /*{{{*/
 {
    Line *l,*n;
-   Mark *m, *m1;
-
-   while (buf->narrow != NULL) widen_buffer(buf);
+   Jed_Mark_Array_Type *m, *m1;
+   
+   jed_widen_whole_buffer (buf);
    if (buf -> beg != NULL) for (l = buf -> beg; l != NULL; l = n)
      {
 	n = l -> next;
 	free_line(l);
 	/* SLFREE( l->data); SLFREE( l); */
      }
-   m = buf->marks;
-   while (m != NULL)
-      {
-	 m1 = m->next;
-	 if (m == &buf->mark_array[buf->mark_ptr - 1])
-	   {
-	      buf->mark_ptr--;
-	   }
-	 else SLFREE(m);
-	 m = m1;
-      }
-   m = buf->spots;
+
+   m = buf->mark_array;
    while (m != NULL)
      {
 	m1 = m->next;
-	if (m == &buf->spot_array[buf->spot_ptr - 1])
-	  {
-	     buf->spot_ptr--;
-	  }
-	else SLFREE(m);
+	SLFREE (m);
+	m = m1;
+     }
+   
+   m = buf->spot_array;
+   while (m != NULL)
+     {
+	m1 = m->next;
+	SLFREE (m);
 	m = m1;
      }
    
    if (buf->user_marks != NULL) free_user_marks (buf);
-
-   if (buf->undo != NULL) delete_undo_ring(buf);
-#ifdef COLOR_COLUMNS
+   
+   if (buf->undo != NULL) delete_undo_ring (buf);
+   
+#if JED_HAS_BUFFER_LOCAL_VARS 
+   jed_delete_blocal_vars (buf->blocal_table);
+#endif
+   
+#if JED_HAS_SAVE_NARROW
+   while (buf->save_narrow != NULL)
+     jed_free_saved_narrow ();
+#endif
+   
+#if JED_HAS_COLOR_COLUMNS
    if (buf->column_colors != NULL) SLFREE (buf->column_colors);
 #endif
    buf->prev->next = buf->next;
    buf->next->prev = buf->prev;
-
+   
+#if JED_HAS_SUBPROCESSES
+   if (buf->subprocess) jed_kill_process (buf->subprocess - 1);
+#endif
+   
    SLFREE( buf);
 }
 
-/* need to make this exploit MaxLine info! */
-void goto_line(int *np)
+/*}}}*/
+
+void goto_line (int *np) /*{{{*/
 {
    unsigned int n;
    unsigned int half1 = LineNum / 2;
@@ -769,7 +844,7 @@ void goto_line(int *np)
    
    if (n < LineNum)
      {
-	if (n > half1) 
+	if (n > half1)
 	  {
 	     sn = (int) (LineNum - n);
 	     prevline(&sn);
@@ -792,62 +867,71 @@ void goto_line(int *np)
 	sn = (int) (Max_LineNum - n);
 	prevline(&sn);
      }
+   Point = 0;
 }
 
-Line *dup_line(Line *l)
-{
-    Line *neew;
-    int n;
-    unsigned char *p, *q;
+/*}}}*/
 
+Line *dup_line(Line *l) /*{{{*/
+{
+   Line *neew;
+   int n;
+   unsigned char *p, *q;
+   
 #ifdef KEEP_SPACE_INFO
    n = l->space;
 #else
    n = l->len;
 #endif
    
-    neew = (Line *) SLMALLOC(sizeof(Line));
-    if ((neew == NULL) ||
-        (NULL == (neew->data = (unsigned char *) SLMALLOC(n))))
-      {
-          exit_error("Malloc Error in dup_line.", 0);
-      }
-    neew->next = l->next;
-    neew->prev = l->prev;
-    neew->len = n;
+   neew = (Line *) jed_malloc (sizeof(Line));
+   if ((neew == NULL) ||
+       (NULL == (neew->data = (unsigned char *) jed_malloc (n))))
+     {
+	exit_error("Malloc Error in dup_line.", 0);
+     }
+   neew->next = l->next;
+   neew->prev = l->prev;
+   neew->len = n;
 #ifdef KEEP_SPACE_INFO
    neew->space = l->space;
 #endif
-    p = neew->data;
-    q = l->data;
-
-    while (n--) *p++ = *q++;
-    return(neew);
+   p = neew->data;
+   q = l->data;
+   
+   while (n--) *p++ = *q++;
+   return(neew);
 }
 
-Buffer *find_buffer(char *name)
-{
-    Buffer *b;
+/*}}}*/
 
-    b = CBuf;
-    do
-      {
-	  if (!strcmp(b->name, name)) return(b);
-	  b = b->next;
-      }
-    while(b != CBuf);
+Buffer *find_buffer(char *name) /*{{{*/
+{
+   Buffer *b;
+   
+   b = CBuf;
+   do
+     {
+	if (!strcmp(b->name, name)) return(b);
+	b = b->next;
+     }
+   while(b != CBuf);
    
 #ifdef __os2__
    b = CBuf;
-   do {
-      if (!strcmpi(b->name, name)) return(b);
-      b = b->next;
-   } while(b != CBuf);
+   do
+     {
+	if (!strcmpi(b->name, name)) return(b);
+	b = b->next;
+     }
+   while(b != CBuf);
 #endif
-    return(NULL);
+   return(NULL);
 }
 
-int buffer_exists(Buffer *b)
+/*}}}*/
+
+int buffer_exists(Buffer *b) /*{{{*/
 {
    Buffer *c = CBuf;
    do
@@ -859,37 +943,39 @@ int buffer_exists(Buffer *b)
    return 0;
 }
 
-int switch_to_buffer(Buffer *buf)
+/*}}}*/
+
+int switch_to_buffer(Buffer *buf) /*{{{*/
 {
-   /* unlink buf */
    /*  save this buffer position */
    CBuf->line = CLine;
    CBuf->point = Point;
    CBuf->linenum = LineNum;
    CBuf->max_linenum = Max_LineNum;
+   
    /* local variables */
-   CBuf->tab = Buffer_Local.tab;
-   CBuf->sd = Buffer_Local.sd;
+   SLMEMCPY ((char *) &CBuf->local_vars, (char *) &Buffer_Local, sizeof (Buffer_Local_Type));
    
    if (buf == CBuf) return(0);
+   
+#if 0
    buf->prev->next = buf->next;
    buf->next->prev = buf->prev;
-
+   
    buf->next = CBuf;
    buf->prev = CBuf->prev;
    CBuf->prev->next = buf;
    CBuf->prev = buf;
-
-
+#endif
+   
    /* now restore new buffer */
    CBuf = buf;
    CLine = CBuf->line;
    Point = CBuf->point;
    
    /* Buffer local variables */
-   Buffer_Local.tab = CBuf->tab;
-   Buffer_Local.sd = CBuf->sd;
-   
+   SLMEMCPY ((char *) &Buffer_Local, (char *) &CBuf->local_vars, sizeof (Buffer_Local_Type));
+	   
    LineNum = CBuf->linenum;
    Max_LineNum = CBuf->max_linenum;
    
@@ -903,39 +989,68 @@ int switch_to_buffer(Buffer *buf)
    return(1);
 }
 
+/*}}}*/
+
 /* we assume that file has already been expanded--- that is, canonical. */
-Buffer *find_file_buffer(char *file)
+Buffer *find_file_buffer(char *file) /*{{{*/
 {
    Buffer *b;
    char *f;
    long n, m;
+#ifdef REAL_UNIX_SYSTEM
+   int inode, device;
+   
+   get_inode_info (file, &device, &inode);
+#endif
    
    f = extract_file(file);
    n = f - file;
    
    b = CBuf;
-    do
-      {
-	 m = strlen(b->dir);
-#ifndef __os2__
-	 if ((m == n) && (!strcmp(b->file,f))
-	     && (!strncmp(b->dir, file, (int) n)))
+   do
+     {
+#ifdef REAL_UNIX_SYSTEM
+	if ((device != -1) && (inode != -1)
+	    && (b->device != -1) && (b->inode != -1))
+	  {
+	     if ((inode == b->inode) 
+		 && (device == b->device)
+		 && !strcmp(b->file, f))
+	       return b;
+	  }
+	else 
+	  {
+	     m = strlen(b->dir);
+	     if ((m == n) && (!strcmp(b->file,f))
+		 && (!strncmp(b->dir, file, (unsigned int) n)))
+	       return b;
+	  }
 #else
- 	 if ((m == n) && (!strcmpi(b->file,f))
- 	     && (!strncmpi(b->dir, file, (int) n)))
+	m = strlen(b->dir);
+	
+# ifndef __os2__
+	if ((m == n) && (!strcmp(b->file, f))
+	    && (!strncmp(b->dir, file, (int) n)))
+# else
+	  if ((m == n) && (!strcmpi(b->file,f))
+	      && (!strncmpi(b->dir, file, (int) n)))
+# endif
+	    return(b);
 #endif
-	   return(b);
-          b = b->next;
-      }
-    while (b != CBuf);
-
-    return(NULL);
+	b = b->next;
+     }
+   while (b != CBuf);
+   
+   return(NULL);
 }
 
+/*}}}*/
+
 /* take a dir and a filename, expand them then put in buffer structure */
-void buffer_filename(char *dir, char *file)
+void buffer_filename(char *dir, char *file) /*{{{*/
 {
-   strcpy(CBuf->dir, dir);
+   strcpy (CBuf->dir, dir);
+   
    if ((file == NULL) || (*file == 0))
      {
 	file = extract_file(CBuf->dir);
@@ -943,15 +1058,21 @@ void buffer_filename(char *dir, char *file)
 	*file = 0;
      }
    else strcpy(CBuf->file, file);
+   
+#ifdef REAL_UNIX_SYSTEM
+   get_inode_info (CBuf->dir, &CBuf->device, &CBuf->inode);
+#endif
    uniquely_name_buffer(CBuf->file);
 }
 
+/*}}}*/
+
 
 /* This kills all undo information! */
-int erase_buffer()
+int erase_buffer() /*{{{*/
 {
    Line *beg, *tthis;
-
+   
    /* CLine = CBuf->end; */
    beg = CBuf->beg;
    bob();
@@ -962,7 +1083,7 @@ int erase_buffer()
 	tthis = CLine;
 	beg->next = tthis->next;
 	tthis->prev = beg;
-
+	
 	update_marks(LDELETE, 1);
 	CLine = tthis->next;
 	Max_LineNum--;
@@ -981,10 +1102,12 @@ int erase_buffer()
    return(1);
 }
 
-void mark_buffer_modified(int *flag)
+/*}}}*/
+
+void mark_buffer_modified(int *flag) /*{{{*/
 {
    unsigned long now;
-
+   
    if (*flag)
      {
 	if (CBuf->flags & BUFFER_TRASHED) return;
@@ -996,7 +1119,9 @@ void mark_buffer_modified(int *flag)
    else CBuf->flags &= ~BUFFER_TRASHED;
 }
 
-void check_line()
+/*}}}*/
+
+void check_line() /*{{{*/
 {
    if ((CLine->len == 1) && (*CLine->data == '\n') && (CLine->data != NewLine_Buffer))
      {
@@ -1007,3 +1132,5 @@ void check_line()
 #endif
      }
 }
+
+/*}}}*/
