@@ -1,19 +1,47 @@
+/* -*- mode: C; mode: fold; -*- */
 /*
  *  Copyright (c) 1992, 1995 John E. Davis  (davis@space.mit.edu)
  *  All Rights Reserved.
  */
+#include "config.h"
+#include "jed-feat.h"
+
+/*{{{ system include files */
+
+#ifdef __WIN32__
+/* This needs to go first before stdio is included. */
+# include <windows.h>
+# define sleep Sleep
+#endif
+
 #include <stdio.h>
+#include <slang.h>
+
+#include "jdmacros.h"
+
 #include <string.h>
 #include <limits.h>
 
+#ifdef HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
 
-#ifdef unix
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
+#ifdef __unix__
 # include <sys/types.h>
 # include <sys/stat.h>
 # include <sys/file.h>
-# ifdef SYSV
-#   include <sys/fcntl.h>
-# endif
+#endif
+
+#ifdef REAL_UNIX_SYSTEM
+# include <utime.h>
+#endif
+
+#ifdef HAVE_SYS_FCNTL_H
+# include <sys/fcntl.h>
 #endif
 
 #ifdef __os2__
@@ -36,11 +64,13 @@
 #include <fcntl.h>
 #endif
 
-#ifdef __DECC
-#include <unixio.h>
-#include <unixlib.h>
+#if defined(__DECC) && defined(VMS)
+# include <unixio.h>
+# include <unixlib.h>
 #endif
 
+#include <errno.h>
+
 /* Was anything missed? */
 #ifndef O_RDONLY
 #ifdef VMS
@@ -50,23 +80,31 @@
 #endif
 #endif
 
-#include "config.h"
+
+/*}}}*/
+/*{{{ local inclue files */
+
 #include "buffer.h"
 #include "file.h"
 #include "misc.h"
 #include "sysdep.h"
 #include "paste.h"
-#include "slang.h"
 #include "ins.h"
 #include "ledit.h"
 
+/*}}}*/
+
 #if defined (msdos) || defined (__os2_16__)
 #define MAX_LINE_LEN 1024
 #else
 #define MAX_LINE_LEN 64 * 1024
 #endif
 
+int Jed_Backup_By_Copying = 0;
+
 #ifdef VMS
+/*{{{ vms_stupid_open */
+
 static int vms_max_rec_size;
 #include <stat.h>
 #include <rms.h>
@@ -130,18 +168,20 @@ int vms_stupid_open(char *file)
    if (ret >= 0) chmod(file, mode);
    return ret;
 }
+
+/*}}}*/
 #endif
 
 int Require_Final_Newline = 0;
 
 /* 0 = read, 1 = write , 2 = append... */
-static int sys_open(char *file, int acces)
+static int sys_open(char *file, int acces) /*{{{*/
 {
    int fp = -1;
    int flags;
    unsigned int mode;
 #ifdef VMS
-   char *p, neew[256];
+   char *p, neew[JED_MAX_PATH_LEN];
 #endif
    
 #ifdef O_BINARY
@@ -190,12 +230,9 @@ static int sys_open(char *file, int acces)
    return(fp);
 }
 
-/* Leaves Point at last line inserted */
-/* returns -1 if unable to open file,
-           -2 if memory allocation error
-           otherwise returns number of lines read */
+/*}}}*/
 
-char *file_type(char *file)
+char *file_type(char *file) /*{{{*/
 {
    char *p, *psave;
    if ((file == (char *) NULL) || (*file == 0)) return( (char *) NULL);
@@ -208,7 +245,8 @@ char *file_type(char *file)
    if (p == file) return(psave); else return(p);
 }
 
-void set_file_modes()
+/*}}}*/
+void set_file_modes () /*{{{*/
 {
    char *type;
 
@@ -263,7 +301,11 @@ void set_file_modes()
    else CBuf->modes = NO_MODE;
 }
 
-int read_file(char *file)
+/*}}}*/
+
+/*{{{ reading/inserting files */
+
+int read_file(char *file) /*{{{*/
 {
    int fp;
    int n, status;
@@ -284,7 +326,8 @@ int read_file(char *file)
    return n;
 }
 
-int insert_file_pointer(VFILE *vp)
+/*}}}*/
+int insert_file_pointer(VFILE *vp) /*{{{*/
 {
    int n = 0;
    unsigned int num;
@@ -300,7 +343,8 @@ int insert_file_pointer(VFILE *vp)
    return(n);
 }
 
-int insert_file(char *file)
+/*}}}*/
+int insert_file(char *file) /*{{{*/
 {
    VFILE *vp;
    int n;
@@ -311,10 +355,13 @@ int insert_file(char *file)
    return(n);
 }
 
+/*}}}*/
 
-#define WRITE jed_write1
+/*}}}*/
 
-#ifdef unix
+/*{{{ writing to files */
+
+#ifdef __unix__
 # define BUFSIZE 0x10000
 #else
 #ifdef VMS
@@ -331,9 +378,9 @@ static char *Output_Bufferp_max;
 
 
 /* definitely perform the write.  Return number of chars written */
-static int jed_write1(int fd, char *b, unsigned int n)
+static int jed_write1(int fd, char *b, unsigned int n) /*{{{*/
 {
-#if !defined(msdos) || defined(__WATCOMC__)
+#if !defined(msdos) || defined(__WATCOMC__) || defined(__WIN32__)
    int len;
    unsigned int total = 0;
 #ifdef VMS
@@ -356,10 +403,28 @@ static int jed_write1(int fd, char *b, unsigned int n)
 	if (bmax == b) bmax = b + (len - 1); /* cannot be helped */
 	len = (int) (bmax - b) + 1;
 #endif
-	if ((len = write (fd, b, len)) < 0)
+	while (-1 == (len = write (fd, b, len)))
 	  {
-	     /* If the write is interruped, what should I do?? */
-	     break;
+#ifdef EINTR
+	     if (errno == EINTR)
+	       continue;
+#endif
+#ifdef EAGAIN
+	     if (errno == EAGAIN)
+	       {
+		  sleep (1);
+		  continue;
+	       }
+#endif
+#ifdef ENOSPC
+	     if (errno == ENOSPC)
+	       {
+		  msg_error ("Write Failed: Disk Full.");
+		  return total;
+	       }
+#endif
+	     msg_error ("Write Failed: Unknown Error.");
+	     return total;
 	  }
 	
 	total += (unsigned int) len;
@@ -382,36 +447,78 @@ static int jed_write1(int fd, char *b, unsigned int n)
 #endif
 }
 
+/*}}}*/
 
 
 /* RMS wants to start a NEW record after a write so just forget it! */
 /* maybe do write-- return number of chars possibly written */
-
-static int jed_write(int fd, char *b, unsigned int n)
+static int jed_write(int fd, char *b, unsigned int n) /*{{{*/
 {
    int num, max, nl_flag = 0;
    unsigned int nsave = n;
    int cr_flag = CBuf->flags & ADD_CR_ON_WRITE_FLAG;
-
+   
+#ifdef MAP_CR_TO_NL_FLAG
+   if (CBuf->flags & MAP_CR_TO_NL_FLAG)
+     {
+	char *bmax = b + n;
+	char *p, *pmax, ch;
+	p = Output_Bufferp;
+	pmax = Output_Bufferp_max;
+	
+	while (b < bmax)
+	  {
+	     ch = *b++;
+	     if ((ch == '\r') || (ch == '\n'))
+	       {
+		  if (cr_flag)
+		    {
+		       *p++ = ch;
+		       if (p == pmax)
+			 {
+			    num = (int) (Output_Bufferp_max - Output_Bufferp);
+			    if (num != jed_write1 (fd, Output_Bufferp, num))
+			      return -1;
+			    Output_Bufferp = Output_Buffer;
+			    p = Output_Bufferp;
+			 }
+		    }
+		  *p++ = '\n';
+	       }
+	     else *p++ = ch;
+	     
+	     if (p == pmax)
+	       {
+		  num = (int) (Output_Bufferp_max - Output_Bufferp);
+		  if (num != jed_write1 (fd, Output_Buffer, num))
+		    return -1;
+		  Output_Bufferp = Output_Buffer;
+		  p = Output_Bufferp;
+	       }
+	  }
+	Output_Bufferp = p;
+	return nsave;
+     }
+#endif		  
    /* amount of space left in buffer */
    /* copy whats in b to the output buffer */
    while (n > 0)
      {
 	num = (int) (Output_Bufferp_max - Output_Bufferp);
-	if (n > num)
+	if ((int) n > num)
 	  {
 #ifdef VMS
 	     max = (int) (Output_Bufferp - Output_Buffer);
 	     if (max)
 	       {
-		  if (max != WRITE(fd, Output_Buffer, max))
+		  if (max != jed_write1(fd, Output_Buffer, max))
 		    return(-1);
 		  Output_Bufferp = Output_Buffer;
 		  continue;
 	       }
 #endif		  
 	     max = num;
-	     MEMCPY(Output_Bufferp, b, max);
+	     SLMEMCPY(Output_Bufferp, b, max);
 	     Output_Bufferp += max;
 	  }
 
@@ -419,25 +526,25 @@ static int jed_write(int fd, char *b, unsigned int n)
 		 (*(b + (n - 1)) == '\n') && (VFile_Mode == VFILE_TEXT))
 	  {
 	     max = n - 1;
-	     MEMCPY(Output_Bufferp, b, max);
+	     SLMEMCPY(Output_Bufferp, b, max);
 	     Output_Bufferp += max;
 	     *Output_Bufferp++ = '\r';
 	     max++;
 
 	     /* can only write the \r */
-	     if (n == num) nl_flag = 1; else *Output_Bufferp++ = '\n';
+	     if (n == (unsigned int) num) nl_flag = 1; else *Output_Bufferp++ = '\n';
 	  }
 	else
 	  {
 	     max = n;
-	     MEMCPY(Output_Bufferp, b, max);
+	     SLMEMCPY(Output_Bufferp, b, max);
 	     Output_Bufferp += max;
 	  }
 	
 	if (Output_Bufferp == Output_Bufferp_max)
 	  {
 	     Output_Bufferp = Output_Buffer;
-	     if (Output_Buffer_Size != WRITE(fd, Output_Buffer, Output_Buffer_Size)) return(-1);
+	     if (Output_Buffer_Size != jed_write1(fd, Output_Buffer, Output_Buffer_Size)) return(-1);
 	     if (nl_flag)
 	       {
 		  nl_flag = 0;
@@ -450,8 +557,10 @@ static int jed_write(int fd, char *b, unsigned int n)
    return(nsave);
 }
 
+/*}}}*/
+
 /* returns -1 on failure */
-int write_region_to_fp(int fp)
+int write_region_to_fp(int fp) /*{{{*/
 {
    register int pnt, len;
    register Line *first, *last;
@@ -518,7 +627,7 @@ int write_region_to_fp(int fp)
    /* Now flush output buffer if necessary */
    
    len = (int) (Output_Bufferp - Output_Buffer);
-   if (!SLang_Error && len) if (len != WRITE(fp, Output_Buffer, len))
+   if (!SLang_Error && len) if (len != jed_write1(fp, Output_Buffer, len))
      {
 	msg_error(err);
      }
@@ -532,42 +641,71 @@ int write_region_to_fp(int fp)
    return(n);
 }
 
-/* write current buffer to open file pointer. Return number of lines */
+/*}}}*/
 
-int write_region(char *file)
+static int jed_close (int fp) /*{{{*/
+{
+   while (-1 == close(fp))
+     {
+#ifdef EINTR
+#ifndef pc_system
+	if (errno == EINTR) 
+	  {
+	     errno = 0;
+	     sleep (1);
+	     continue;
+	  }
+#endif
+#endif
+	msg_error ("Error closing file.  File system may be full.");
+	return -1;
+     }
+   return 0;
+}
+
+/*}}}*/
+
+int write_region (char *file) /*{{{*/
 {
    int fp;
    int n;
-   char msg[256];
 
    if (!check_region(&Number_Zero)) return(-1);
    if ((fp = sys_open(file, 1)) < 0)
      {
-	sprintf(msg, "Unable to open %s for writing.", file);
-	msg_error(msg);
-	return(-1);
+	jed_verror ("Unable to open %s for writing.", file);
+	return -1;
      }
    n = write_region_to_fp(fp);
-   close(fp);
+   if (-1 == jed_close (fp)) n = -1;
 
    return(n);
 }
 
+/*}}}*/
 
 
 /* returns -1 on failure and number of lines on success */
-
-static int write_file(char *file)
+static int write_file(char *file) /*{{{*/
 {
    Mark *m;
    int n = -1;
    int fnl;
+      
+#ifdef VMS
+   register Line *l;
+   register int len = 0, max = 0;
+#endif
    
+   push_spot();
+   
+#if JED_HAS_SAVE_NARROW
+   jed_push_narrow ();
+   jed_widen_whole_buffer (CBuf);
+#endif
    
 #ifdef VMS
-   register Line *l = CBuf->beg;
-   register int len = 0, max = 0;
-   
+   l = CBuf->beg;
    while (l != NULL)
      {
 	len = l->len;
@@ -577,7 +715,6 @@ static int write_file(char *file)
    vms_max_rec_size = max;
 #endif
    
-   push_spot();
    bob();
    push_mark();  m = CBuf->marks;
    eob();
@@ -597,40 +734,50 @@ static int write_file(char *file)
    Require_Final_Newline = fnl;
    VFile_Mode = VFILE_TEXT;
    if (m == CBuf->marks) pop_mark(&Number_Zero);
+   
+#if JED_HAS_SAVE_NARROW
+   jed_pop_narrow ();
+#endif   
+   
    pop_spot();
    return(n);
 }
 
-int append_to_file(char *file)
+/*}}}*/
+
+int append_to_file(char *file) /*{{{*/
 {
    int fp;
    int n;
 
    if ((fp = sys_open(file, 2)) < 0) return(-1);
    n = write_region_to_fp(fp);
-   close(fp);
+   if (-1 == jed_close (fp)) n = -1;
    check_buffers();
    return(n);
 }
 
-static int make_autosave_filename(char *save, char *dir, char *file)
+/*}}}*/
+
+/*}}}*/
+
+static int make_autosave_filename(char *save, char *dir, char *file) /*{{{*/
 {
    char *s;
-   int dat;
 
    if (*file == 0) return(0);
       
    
    if (SLang_run_hooks("make_autosave_filename", dir, file) && !SLang_Error)
      {
-	if (SLang_pop_string(&s, &dat)) return(0);
-	strncpy(save, s, 254);
-	save[255] = 0;
-	if (dat == 1) SLFREE(s);
+	if (SLpop_string(&s)) return(0);
+	strncpy(save, s, JED_MAX_PATH_LEN - 2);
+	save[JED_MAX_PATH_LEN - 1] = 0;
+	SLFREE(s);
      }
    else
      {
-#ifdef unix
+#ifdef __unix__
 	sprintf(save, "%s#%s#", dir, file);
 #else
 #if defined (msdos) || defined (__os2__)
@@ -643,15 +790,110 @@ static int make_autosave_filename(char *save, char *dir, char *file)
    return(1);
 }
 
-int write_file_with_backup(char *dir, char *file)
+/*}}}*/
+
+#ifndef VMS
+# ifdef REAL_UNIX_SYSTEM
+int jed_copy_file (char *from, char *to) /*{{{*/
 {
-   char neew[256]; char save[256];
+   int mode;
+   short uid, gid;
+   FILE *fp0, *fp1;
+   char buf[0x7FFF];
+   unsigned int readlen;
+   int ret;
+   struct stat st;
+#ifndef NeXT
+   struct utimbuf ut;
+#endif
+
+   if (1 != sys_chmod (from, 0, &mode, &uid, &gid))
+     return -1;		       /* from does not exist as regular file */
+   
+   /* Need file modification times so that they can be preserved. */
+   if (-1 == stat (from, &st))
+     return -1;
+   
+   fp0 = fopen (from, "rb");
+   if (fp0 == NULL) return -1;
+   
+   fp1 = fopen (to, "wb");
+   if (fp1 == NULL) 
+     {
+	(void) fclose (fp0);
+	return -1;
+     }
+   
+   (void) chmod (to, 0600);
+
+   ret = 0;
+   do
+     {
+	readlen = fread (buf, 1, sizeof(buf), fp0);
+	if (readlen)
+	  {
+	     if (readlen != fwrite (buf, 1, readlen, fp1))
+	       {
+		  ret = -1;
+		  break;
+	       }
+	  }
+     }
+   while (readlen == sizeof (buf));
+   
+   fclose (fp0);
+
+   if (EOF == fclose (fp1))
+     {
+	ret = -1;
+     }
+   
+   (void) sys_chmod (to, 1, &mode, &uid, &gid);
+
+#ifndef NeXT
+   /* Set file modification times */
+   ut.actime = st.st_atime;
+   ut.modtime = st.st_mtime;
+   (void) utime (to, &ut);
+#endif
+
+   return ret;
+}
+/*}}}*/
+# endif
+#endif
+
+#ifndef VMS   
+static int perform_backup (char *from, char *to, int try_force_rename) /*{{{*/
+{
+   int ret = -1;
+   
+   if (try_force_rename
+       || (Jed_Backup_By_Copying == 0))
+     {
+	(void) unlink(to);
+	ret = rename (from, to);
+     }
+#ifdef REAL_UNIX_SYSTEM
+   /* Needs porting to OS/2 and DOS */
+   if (ret == -1)
+     ret = jed_copy_file (from, to);
+#endif
+
+   return ret;
+}
+/*}}}*/
+#endif
+
+int write_file_with_backup(char *dir, char *file) /*{{{*/
+{
+   char neew[JED_MAX_PATH_LEN]; char autosave_file[JED_MAX_PATH_LEN];
    int n;
    int mode, do_mode;
    short uid, gid;
 #ifndef VMS
-   char *old;
-   int ok = 0, bu, do_free = 0;
+   char *old = NULL;
+   int backup_went_ok;
 #endif
 #ifdef __os2__
    PHOLDFEA EAs;
@@ -661,34 +903,41 @@ int write_file_with_backup(char *dir, char *file)
 
    sprintf(neew, "%s%s", dir, file);
    
-   if (((CBuf->flags & AUTO_SAVE_BUFFER) == 0)
-       || !make_autosave_filename(save, dir, file)) *save = 0;
+   *autosave_file = 0;
+   if (CBuf->flags & AUTO_SAVE_BUFFER)
+      (void) make_autosave_filename(autosave_file, dir, file);
 
    do_mode = sys_chmod(neew, 0, &mode, &uid, &gid);
-   if ((do_mode < 0) ||  (do_mode > 1)) return(-1);
+   if ((do_mode != 0) && (do_mode != 1)) return -1;
+   /* IF do_mode == 0, then the file does not exist.  This means that 
+    * there is nothing to backup.  If do_mode == 1, the file is a regular
+    * file which we do want to backup.
+    */
+   
 #ifndef VMS
    
-#ifdef __os2__
+# ifdef __os2__
    EAs = QueryEAs (neew);
-#endif
+# endif
    
-   bu = 1;
-   if (((CBuf->flags & NO_BACKUP_FLAG) == 0)
+   backup_went_ok = 0;
+   if (do_mode
+       && ((CBuf->flags & NO_BACKUP_FLAG) == 0)
        && SLang_run_hooks("make_backup_filename", dir, file))
      {
-	if (((bu = !SLang_pop_string(&old, &do_free)) != 0) && (*old))
+	if ((0 == SLpop_string(&old))
+	    && (*old != 0))
 	  {
-	     unlink(old);
-	     ok = !rename(neew, old);
+	     backup_went_ok = (0 == perform_backup (neew, old, 0));
 	  }
      }
-#endif
+#endif				       /* NOT VMS */
    
    if (-1 != (n = write_file(neew)))
      {
-	if (*save)
+	if (*autosave_file)
 	  {
-	     sys_delete_file(save);
+	     sys_delete_file(autosave_file);
 	  }
 	     
 	if (do_mode) /* must be an existing file, so preserve mode */
@@ -704,28 +953,45 @@ int write_file_with_backup(char *dir, char *file)
 	  }
 	/* This is for NFS time problems */
 	CBuf->c_time = sys_file_mod_time(neew);
+	
+	/* Since we wrote the buffer to the file, it is not modified. */
+	if (
+#ifdef __os2__
+	    !strcmpi(file, CBuf->file) && !strcmp(dir, CBuf->dir)
+#else
+	    !strcmp(file, CBuf->file) && !strcmp(dir, CBuf->dir)
+#endif
+	    )
+	  CBuf->flags &= ~FILE_MODIFIED;
+	
 	/* CBuf->c_time = sys_time(); */
      }
 #ifndef VMS
-   else if (ok && bu && *old) rename(old, neew); /* error -- put it back */
-   if ((do_free == 1) && bu) SLFREE(old);
+   /* error -- put it back */
+   else if (backup_went_ok) perform_backup (old, neew, 1); 
+   
+   if (old != NULL) SLFREE(old);
 #endif
    return(n);
 }
 
-/* warning-- this saves on the narrowed part of buffer.
-   Here, I widen first.  I need a save_restriction type of thing because
-   I do not narrow back.
-   */
-void auto_save_buffer(Buffer *b)
+/*}}}*/
+
+void auto_save_buffer(Buffer *b) /*{{{*/
 {
-   char tmp[256];
+   char tmp[JED_MAX_PATH_LEN];
    Buffer *old_buf;
    unsigned int vfm = VFile_Mode;
    
    if (b == NULL) return;
    b->hits = 0;
-   if ((b->narrow != NULL) || !(b->flags & BUFFER_TRASHED)) return;
+   if ((0 == (b->flags & BUFFER_TRASHED))
+       || (0 == (b->flags & (AUTO_SAVE_BUFFER | AUTO_SAVE_JUST_SAVE))))
+     return;
+
+#if !JED_HAS_SAVE_NARROW
+   if (b->narrow != NULL) return;
+#endif
 
    old_buf = CBuf;
    switch_to_buffer(b);
@@ -740,45 +1006,53 @@ void auto_save_buffer(Buffer *b)
 	     flush_message("autosaving..."); 
 	     sys_delete_file(tmp);
 	     write_file(tmp);
+#ifndef pc_system
+	     (void) chmod (tmp, 0600);
+#endif
 	     message("autosaving...done");
 	  }
       }
-   else if (b->flags & AUTO_SAVE_JUST_SAVE)
+   else if (write_file_with_backup(b->dir, b->file) >= 0)
      {
-	if (write_file_with_backup(b->dir, b->file) >= 0)
-	  {
-	     b->flags &= ~BUFFER_TRASHED;
-	  }
+	b->flags &= ~BUFFER_TRASHED;
      }
+
    switch_to_buffer(old_buf);
    VFile_Mode = vfm;
 }
 
-void auto_save_all()
+/*}}}*/
+
+void auto_save_all() /*{{{*/
 {
     Buffer *b;
 
    if (NULL == (b = CBuf)) return;
    do
      {
-	while (b->narrow != NULL) widen_buffer(b);
+	jed_widen_whole_buffer (b);
 	if ((*b->file != 0) && (b->hits != 0)) auto_save_buffer(b);
 	b = b->next;
      }
    while (b != CBuf);
 }
 
-#ifdef unix
-#ifndef __GO32__
-static int is_link(char *f, char *f1)
+/*}}}*/
+
+#ifdef REAL_UNIX_SYSTEM
+/*{{{ symbolic link stuff */
+
+static int is_link(char *f, char *f1) /*{{{*/
 {
    struct stat s;
-   int l;
+   unsigned int l;
    int is_dir = 0;
-   char work[256];
+   char work[JED_MAX_PATH_LEN];
    
-   l = strlen(f) - 1;
-   if ((f[l] == '/') && (l > 1))
+   l = strlen(f);
+   if (l == 0) return 0;
+   l--;
+   if ((l > 1) && (f[l] == '/'))
      {
 	strcpy(work, f);
 	is_dir = 1;
@@ -792,7 +1066,7 @@ static int is_link(char *f, char *f1)
        || (((s.st_mode & S_IFMT) & S_IFLNK) == 0))
      return(0);
    
-   l = 255;
+   l = JED_MAX_PATH_LEN - 1;
    if (is_dir)
      {
 	/* This way, it will be expanded properly upon return */
@@ -800,34 +1074,78 @@ static int is_link(char *f, char *f1)
 	l -= 4;			       /* 1 for final slash */
      }
 	     
-   if (-1 == (l = readlink(f, f1, l))) return(0);
+   if (-1 == (int) (l = readlink(f, f1, l))) return 0;
    if (is_dir) f1[l++] = '/';
    f1[l] = 0;
+   
+   /* If the link is an absolute pathname (starts with /) then the 
+    * ../ prefix should not go there.
+    */
+   if (*f1 == '/')
+     {
+	char ch;
+	
+	f = f1 - 3;
+	while (0 != (ch = *f1++))
+	  *f++ = ch;
+	
+	*f = 0;
+     }
+
    return(1);
 }
 
-/* we do not worry about the dir--- only the filename */
-char *expand_link(char *f)
+/*}}}*/
+
+/* This routine should be modified to work with all components of the path
+ * and not just the last component.  For example, suppose /usr/lib/jed is a 
+ * link to /usr/local/src/jed and one tries to read the file
+ * /usr/lib/jed/lib/site.sl.  This should be expanded to 
+ * /usr/local/lib/jed/lib/site.sl.  Currently, it will only have an effect if
+ * site.sl is itself a link.
+ */
+char *expand_link(char *f) /*{{{*/
 {
-   char work[256], lnk[256];
+   char work[JED_MAX_PATH_LEN], lnk[JED_MAX_PATH_LEN];
    char *d = expand_filename(f);
  
   
    if (is_link(d, lnk))
      {
-	strcpy(work, d);
-	*(extract_file(work)) = 0;
-	strcat(work, lnk);
+	char *w, *w1, ch;
+	
+	if (*lnk == '/')	       /* absolute */
+	  strcpy (work, lnk);
+	else
+	  {
+	     strcpy (work, d);
+	     *(extract_file(work)) = 0;
+	     strcat (work, lnk);
+	  }
+	
+	/* Collapse multiple // since expand_filename will interpret them
+	 * differently.
+	 */
+	w = w1 = work;
+	while ((ch = *w1++) != 0)
+	  {
+	     *w++ = ch;
+	     if (ch == '/') while (*w1 == '/') w1++;
+	  }
+	*w = 0;
+	
 	d = expand_filename(work);
      }
    
    return (d);
 }
 
-#endif
+/*}}}*/
+
+/*}}}*/
 #endif
 
-void visit_file(char *dir, char *file)
+void visit_file(char *dir, char *file) /*{{{*/
 {
 #ifndef __os2__
    if (strcmp(file, CBuf->file) || strcmp(dir, CBuf->dir))
@@ -848,8 +1166,9 @@ void visit_file(char *dir, char *file)
    check_buffers();
 }
 
+/*}}}*/
 
-char *dir_file_merge(char *dir, char *file)
+char *dir_file_merge(char *dir, char *file) /*{{{*/
 /* 
  * returns result of merging dir with file. If file is empty, dir is
  * considered to be whole file.
@@ -866,7 +1185,9 @@ char *dir_file_merge(char *dir, char *file)
    return expand_filename(name);
 }
 
-int file_status(char *file)
+/*}}}*/
+
+int file_status(char *file) /*{{{*/
 /*
  *  Returns a coded integer about file.  If the file does not exist, 0 is
  *  returned.  Meaning:
@@ -884,16 +1205,19 @@ int file_status(char *file)
    return sys_chmod(file, 0, &mode, &uid, &gid);
 }
 
-int file_changed_on_disk(char *file)
+/*}}}*/
+int file_changed_on_disk(char *file) /*{{{*/
 {
    unsigned long t;
    Buffer *buf;
    if (NULL == (buf = find_file_buffer(file))) return(0);
+   if (buf->flags & FILE_MODIFIED) return 1;
    t = sys_file_mod_time(file);
    return(t > buf->c_time);
 }
 
-int file_time_cmp(char *file1, char *file2)
+/*}}}*/
+int file_time_cmp(char *file1, char *file2) /*{{{*/
 {
    unsigned long t1, t2;
    
@@ -912,13 +1236,24 @@ int file_time_cmp(char *file1, char *file2)
    return(-1);
 }
 
-void auto_save(void)
+/*}}}*/
+
+void auto_save(void) /*{{{*/
 {
    auto_save_buffer(CBuf);
 }
 
-void check_buffer(Buffer *b)
+/*}}}*/
+
+void check_buffer(Buffer *b) /*{{{*/
 {
+#ifdef REAL_UNIX_SYSTEM
+   /* Update (or even invalidate) inode information since directories may have
+    * been renamed, etc...
+    */
+   (void) get_inode_info (b->dir, &b->device, &b->inode);
+#endif
+
    if ((*b->file != 0)
        && file_changed_on_disk(dir_file_merge(b->dir, b->file)))
      {
@@ -927,7 +1262,8 @@ void check_buffer(Buffer *b)
    else b->flags &= ~FILE_MODIFIED;
 }
 
-void check_buffers()
+/*}}}*/
+void check_buffers() /*{{{*/
 {
    Buffer *b = CBuf;
    do
@@ -938,12 +1274,16 @@ void check_buffers()
    while (b != CBuf);
 }
 
-void set_file_trans(int *mode)
+/*}}}*/
+
+void set_file_trans(int *mode) /*{{{*/
 {
    if (*mode) VFile_Mode = VFILE_BINARY; else VFile_Mode = VFILE_TEXT;
 }
 
-int read_file_pointer(int fp)
+/*}}}*/
+
+int read_file_pointer(int fp) /*{{{*/
 {
    int n = 1;
    unsigned int num;
@@ -959,14 +1299,14 @@ int read_file_pointer(int fp)
 #endif
      remake_line(CLine->len + num + 1);
    
-   MEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
+   SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
    CLine->len = num;
    
    while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
      {
 	n++;
 	if ((num == 1) && (*vbuf == '\n')) make_line(num); else make_line(num + 1);
-	MEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
+	SLMEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
 	CLine->len = num;
 	if (SLang_Error) break;
      }
@@ -979,4 +1319,32 @@ int read_file_pointer(int fp)
    return n;
 }
 
+/*}}}*/
 
+#ifdef REAL_UNIX_SYSTEM
+int get_inode_info (char *dir, int *device, int *inode) /*{{{*/
+{
+   struct stat st;
+   char work[512];
+   char *f;
+   
+   f = extract_file (dir);
+   if (*f != 0)
+     {
+	unsigned int len = f - dir;
+	strncpy (work, dir, len);
+	work[len] = 0;
+	dir = work;
+     }
+   
+   *inode = *device = -1;
+   if (-1 == stat (dir, &st)) return -1;
+   
+   *inode = st.st_ino;
+   *device = st.st_dev;
+   
+   return 0;
+}
+/*}}}*/
+#endif
+   
